/**
 *	tbl_read.c	n.k.	june 9, 1992
 *	tb2_read.c	k.f.	july 29, 1996
 *
 * Modifications:
 *   k.f. sep.25 1996   rewrite whole structure for v2.0
 *   k.f. oct.8  2004   skip null line, check separator,keyword before others
 *
 *  tb2_Spec * tb2_read( char* abbr );
 *        read specified table file into buffer
 *    abbr - abbreviated name of the table
 *           example: "mgv3" seraches for "mgv3tbl.tbl" at default directory
 *    return: pointer to tb2_Spec structure on success
 *            0 -- error (error code in tb2_errno, see tb2.h)
 *
 *  (old interface for compatibility)
 *  int tbl_read( char* abbr )  -  read specified table file into buffer
 *    abbr - abbreviated name of the table
 *           example: "mg" seraches for "mgtbl.tbl" at default directory
 *  return: 0 -- successfully completed
 *         <0 -- error (see tb2.h)
 *
**/
static char what_tb2read[] =
"@(#)tb2_read.c, v2.1.0, n.f, k.f, linac table read, jun.1992-oct.2004";

/*** definitions ***/
#include  <stdio.h>
#include  <stdlib.h>			/* for getenv() */
#include  <string.h>			/* for strtok() */
#include  "tb2.h"			/* error codes, default directory */

/*** other variables ***/
int     tb2_errno;			/* error code holder */
tb2_Spec *default_tbl = (void *)NULL;	/* default table for tbl_xxx */

#ifdef MSDOS
char *getenv();				/* to get env.variable TBLDIR */
#endif

/*** internal function ***/
static char * _tb2_strdupaddn( char * format );

/**
 *   tb2_read() - read specified table into buffer
**/
tb2_Spec *tb2_read( abbr )
char  *abbr;			/* abbr. name of the table */
{
    FILE  *fp;			/* table file pointer */
    char  *tbldir;		/* getenv() pointer */
    char  buff[L_BUFFSIZE+1], buffent[L_KEY*4];
    int   nkeys=0;		/* # of valid lines in the table */
    char  filename[L_FNAME+1];	/* table name holder */
    int   i,j,k,l;
    int   loop, ipos[N_POS];
    char  separator, format[L_FORMAT+1];

    tb2_Spec *tb2spec;				/* database */
    struct _tb2_paramkey (*tb2paramkey)[];	/* pointer to array of keys */
    tb2_Entry *tb2entry;			/* pointer to a entry */
    union _tb2_parament (*tb2parament)[];	
    					/* ptr to array of param of a entry */

/*** clear table buffer ***/
    tb2_errno = 0;

/*** search for the table ***/
    tbldir = getenv("TBLDIR");
    if( tbldir != NULL ) {			/* (a) env. variable */
	if( *(tbldir + strlen(tbldir) - 1 ) == '/' )
	    sprintf( filename, "%s%stbl.tbl", tbldir, abbr );
	else
	    sprintf( filename, "%s/%stbl.tbl", tbldir, abbr );
    } else {				/* (b) default */
	sprintf( filename, "%s%stbl.tbl", DEF_DIR, abbr );
    }
    fp = fopen( filename, "r" );
    if( fp == NULL )  goto EOPEN;

  /* [0] allocate data structure */
    if (!(tb2spec = (tb2_Spec *) tb2_malloc( sizeof (tb2_Spec) )))
	goto EMALLOC;					/* table spec */
    if (!(tb2spec->hash_keyword = tb2_hashcreate('s','i')))
	goto EMALLOC;			/* key (string) index */
    if (!(tb2spec->hash_name = tb2_hashcreate('s','m')))
	goto EMALLOC;			/* name (string) index */
    if (!(tb2spec->hash_serial = tb2_hashcreate('i','m')))
	goto EMALLOC;			/* serial (integer) index */
    if ( !(tb2spec->abbrev = tb2_strdup( abbr )) )
	goto EMALLOC;
    if ( !(tb2spec->filename = tb2_strdup( filename )) )
	goto EMALLOC;
    
/*** read table ***/
  /* [1] initial values */
    loop = 0;
    separator = DEF_SEPARATOR;
    strncpy( format, DEF_FORMAT, L_FORMAT );
    nkeys = DEF_NKEYS;
    tb2spec->num_param = 0;

  /********************************************/
  /* [2] read table loop (while) : start here */
    while( fgets( buff, L_BUFFSIZE, fp ) != NULL ) {
	char * eachtoken;

	i = strlen( buff );
	if (i == 1) continue;		/* ignore null line */
	buff[i-1]='\0';			/* remove new line */

  /* [3] search key's (separator, keywords, format) */
	if( loop == 0 ) {
  /* [3-1] separator */
	    if( !strncmp( buff, "!separa", 7 ) ) {
		i = _tbl_position( buff, '\'', ipos );
		if( i < 0 ) {
		    goto EPOSITION;
		} else if( i >= 2 ) {
		    separator = buff[ ipos[0]+1 ];
#ifdef DEBUG1
		    printf( "    -> separator specified is '%c'\n", 
			   separator );
#endif
		}
		continue;
  /* [3-2] keyword */
	    } else if( !strncmp( buff, "!keywor", 7 ) ) {
		i = _tbl_position( buff, '"', ipos );
		if( i < 0 ) {
		    goto EPOSITION;
		} else if( i == 0 || i == 1 ) {
		    goto EPOSITION;
		}
		/* } else if( i >= 2 ) { */
		buff[ ipos[i-1] ] = '\0';
		j = ipos[0] + 1;    /* offset of keywords entry */
		nkeys = _tbl_position( buff+j, separator, ipos ) + 1;
		if( nkeys <= 0 )  goto EPOSITION;
#ifdef DEBUG1
		printf( "    -> keyword specified is '%s'\n", buff+j );
#endif
		if( nkeys > N_KEYS )  {
		    /* goto ENKEYS; */
		    fprintf( stderr, "warning: number of keywords '%d' is \
greater than '%d'\n", nkeys, N_KEYS );
		}

		/* allocate container structure for keywords */
		tb2paramkey = (struct _tb2_paramkey (*)[]) tb2_malloc(
		    (nkeys + 1) * sizeof (struct _tb2_paramkey) );
		if ( !tb2paramkey )
		    goto EMALLOC;
		tb2spec->tb2_paramkey = tb2paramkey;
			
		for( k=0 ; k<(nkeys-1) ; k++ )
		    buff[ j+ipos[k] ] = '\0';
		
		for( k=0 ; k<nkeys ; k++ ) {
		    switch( k ) {
		    case 0:
			if( strlen( buff+j ) > (size_t)L_KEY ) {
			    goto ELONGKEY;
			}
			(*tb2paramkey)[0].keyword = tb2_strdup( buff+j );
			if ( !(*tb2paramkey)[0].keyword )
			    goto EMALLOC;
			break;

		    default:
			if( strlen( buff+j+ipos[k-1]+1 ) > (size_t)L_KEY ) {
			    j = j+ipos[k-1]+1;
			    goto ELONGKEY;
			}
			(*tb2paramkey)[k].keyword = 
			    tb2_strdup( buff+j+ipos[k-1]+1 );
			if ( !(*tb2paramkey)[k].keyword )
			    goto EMALLOC;
			break;
		    }
		}
		(*tb2paramkey)[nkeys].keyword = (void *)NULL;
		if ( (tb2_keywordhash( tb2spec, tb2paramkey )) )
		    goto EMALLOC;

		tb2spec->num_param = nkeys;

#ifdef DEBUG1
		printf( "       %d keys -- ", nkeys );
		for( k=0 ; k<nkeys ; k++ ) {
		    printf( " %s, ", (*tb2paramkey)[k].keyword );
		}
		printf( "\n" );
#endif
		continue;
  /* [3-3] format */
	    } else if( !strncmp( buff, "!format", 7 ) ) {

		i = _tbl_position( buff, '"', ipos );
		if( i < 0 ) {
		    goto EPOSITION;
		} else if( i == 0 || i == 1 ) {
		    goto EPOSITION;
		}
		/* } else if( i >= 2 ) { */
		buff[ ipos[i-1] ] = '\0';
		strncpy( format, buff+ipos[0]+1, L_FORMAT );
		/* format[L_FORMAT] = '\0'; */
#ifdef DEBUG1
		printf( "    -> format specified is '%s'\n", format );
#endif
		if( strlen( buff+ipos[0]+1 ) > (size_t)L_FORMAT )
		    goto EFORMAT;
		/* copy of the format */
		if ( !(eachtoken = tb2_strdup( format )) )
		    goto EMALLOC;		/* for re-entrant */
#define TOKDELIM " \t"
		/* search first token */
		if ( !(eachtoken = strtok(eachtoken, TOKDELIM)) )
		    goto EFORMAT;
		for( k=0 ; k<nkeys ; k++ ) {
		    /* copy of a format with "%n" added */
		    if ( !(eachtoken = _tb2_strdupaddn( eachtoken )) )
			goto EMALLOC;
		    (*tb2paramkey)[k].format = eachtoken;
		    if ( strstr(eachtoken,"dl") )	/* long */
			(*tb2paramkey)[k].type = 'l';
		    else if ( strstr(eachtoken,"xl") )	/* long */
			(*tb2paramkey)[k].type = 'l';
		    else if ( strstr(eachtoken,"lf") )	/* double */
			(*tb2paramkey)[k].type = 'g';
		    else if ( strstr(eachtoken,"lg") )	/* double */
			(*tb2paramkey)[k].type = 'g';
		    else if ( strstr(eachtoken,"fl") )	/* double */
			(*tb2paramkey)[k].type = 'g';
		    else if ( strstr(eachtoken,"gl") )	/* double */
			(*tb2paramkey)[k].type = 'g';
		    else if ( strchr(eachtoken,'d') )	/* int (decimal) */
			(*tb2paramkey)[k].type = 'd';
		    else if ( strchr(eachtoken,'x') )	/* int (hexadecimal) */
			(*tb2paramkey)[k].type = 'd';
		    else if ( strchr(eachtoken,'f') )	/* float */
			(*tb2paramkey)[k].type = 'f';
		    else if ( strchr(eachtoken,'g') )	/* float */
			(*tb2paramkey)[k].type = 'f';
		    else if ( strchr(eachtoken,'s') )	/* string */
			(*tb2paramkey)[k].type = 's';
		    /* next token */
		    if ( !(eachtoken = strtok((char *)NULL, TOKDELIM)) )
			if (k == nkeys-1)  break;
			else goto EFORMAT;
		}
		(*tb2paramkey)[nkeys].format = (void *)NULL;

		continue;
	    } else if( buff[0] == '!' ) {
		continue;			/* skip comment line */
	    }
	}
	if( !(tb2spec->num_param) ) continue;	/* no separator,keyword yet */

  /* [4] read contents of the table */
	if( buff[0] == '!' )  continue;		/* skip comment line */
#ifdef DEBUG2
	printf( "%3d: %s\n", loop, buff );
#endif
	tb2parament = (union _tb2_parament (*)[]) 
	    tb2_malloc( (nkeys + 1) * sizeof (union _tb2_parament) );
	if ( !tb2parament )  goto EMALLOC;
	
	j = 0;
	for( k=0 ; k<nkeys ; k++ ) {

	    switch( (*tb2paramkey)[k].type )  {
	    case 's':
		if (1 != sscanf( buff+j, (*tb2paramkey)[k].format, 
				buffent, &l ))  goto ELINE;
		(*tb2parament)[k].string = tb2_strdup( buffent );
		if ( strlen(buffent) > (size_t)L_KEY )  {
		    fprintf( stderr, "warning: length of parameter '%s' is \
greater than '%d'\n", buffent, L_KEY );
		}
		break;
	    case 'd':
		if (1 != sscanf( buff+j, (*tb2paramkey)[k].format, 
				&i, &l ))  goto ELINE;
		(*tb2parament)[k].integer = i;
		break;
	    case 'l':
		{
		    long lint;
		    if (1 != sscanf( buff+j, (*tb2paramkey)[k].format, 
				    &lint, &l ))  goto ELINE;
		    (*tb2parament)[k].longinteger = lint;
		}
		break;
	    case 'f':
		{
		    float freal;
		    if (1 != sscanf( buff+j, (*tb2paramkey)[k].format, 
				    &freal, &l ))  goto ELINE;
		    (*tb2parament)[k].real = freal;
		}
		break;
	    case 'g':
		{
		    double greal;
		    if (1 != sscanf( buff+j, (*tb2paramkey)[k].format, 
				    &greal, &l ))  goto ELINE;
		    (*tb2parament)[k].doublereal = greal;
		}
		break;
	    default:
		goto ELINE;
	    }
	    j += l;			/* advance read buffer pointer */
	}
	if ( !(tb2entry = (tb2_Entry *) tb2_malloc( sizeof (tb2_Entry) )) )
	    goto EMALLOC;
	tb2entry->tb2_parament = tb2parament;
	tb2entry->serial = loop + 1;	/* serial number beginning at 1 */
	tb2entry->name = (*tb2parament)[0].string;

#ifdef DEBUG3
	for ( j=0; j< tb2spec->num_param; j++ )  {
	    static char format[L_FORMAT];
	    strcpy( format, ((*(tb2spec->tb2_paramkey))[j]).format );
	    format[ strlen(format) -2 ] = '\0';

	    switch (((*(tb2spec->tb2_paramkey))[j]).type)  {
	    case 's':
		fprintf( stderr, format, 
			((*(tb2entry->tb2_parament))[j]).string );
		break;
	    case 'd':
		fprintf( stderr, format, 
			((*(tb2entry->tb2_parament))[j]).integer );
		break;
	    case 'l':
		fprintf( stderr, format, 
			((*(tb2entry->tb2_parament))[j]).longinteger );
		break;
	    case 'f':
		fprintf( stderr, format, 
			((*(tb2entry->tb2_parament))[j]).real );
		break;
	    case 'g':
		fprintf( stderr, format, 
			((*(tb2entry->tb2_parament))[j]).doublereal );
		break;
	    }
	    fprintf( stderr, ":%c ", ((*(tb2spec->tb2_paramkey))[j]).type );
	}
	fprintf( stderr, "\n" );
#endif /* DEBUG3 */

	if ( (tb2_enterhash( tb2spec, tb2entry )) )  goto EMALLOC;
	
#ifdef DEBUG2
	printf( "   name='%s'\n", tb2entry->name );
#endif
#ifdef DEBUG3
	for ( j=0; j< tb2spec->num_param; j++ )  {
	    tb2_ParamEnt * param;
	    static char format[L_FORMAT];

	    param = tb2_data( tb2spec, tb2entry->name, 
			      ((*(tb2spec->tb2_paramkey))[j]).keyword );

	    strcpy( format, ((*(tb2spec->tb2_paramkey))[j]).format );
	    format[ strlen(format) -2 ] = '\0';

	    switch (((*(tb2spec->tb2_paramkey))[j]).type)  {
	    case 's':
		fprintf( stderr, format, 
			param->string );
		break;
	    case 'd':
		fprintf( stderr, format, 
			param->integer );
		break;
	    case 'l':
		fprintf( stderr, format, 
			param->longinteger );
		break;
	    case 'f':
		fprintf( stderr, format, 
			param->real );
		break;
	    case 'g':
		fprintf( stderr, format, 
			param->doublereal );
		break;
	    }
	    fprintf( stderr, " " );
	}
	fprintf( stderr, "\n" );
#endif /* DEBUG3 */

	loop++;

    }
  /* [8] read table loop (while) : stop here */
  /*******************************************/

    fclose( fp );
#ifdef DEBUG1
    printf( " total valid lines are %d\n", loop );
#endif
    if( loop == 0 )  goto ENULLTBL;
    tb2spec->num_entry = loop;

/*** end of the routine ***/
    return  tb2spec;

/*** errors ***/
EOPEN:
#ifdef TBL_WARN
    fprintf( WARN_DEST,
	    "tb2_read %s: table='%s' not found, or unable to open\n", 
	    TBL_VER, filename );
#endif
    tb2_errno = TBL_E_NOTFOUND;
    return (void *)NULL;

/*
ENKEYS:
#ifdef TBL_WARN
    fprintf( WARN_DEST,
	    "tb2_read %s: too many (=%d) # of keys (>%d) for tbl='%s'\n",
	    TBL_VER, nkeys, N_KEYS, filename );
#endif
    tb2_errno = TBL_E_KEYWORDS;
    return (void *)NULL;
*/

 ELONGKEY:
#ifdef TBL_WARN
    fprintf( WARN_DEST,
	    "tb2_read %s: keyword='%s' is too long(>%d) for tbl='%s'\n", 
	    TBL_VER, buff + j, L_KEY, filename );
#endif
    tb2_errno = TBL_E_KEYWORDS;
    return (void *)NULL;

 EFORMAT:
#ifdef TBL_WARN
    fprintf( WARN_DEST,
	    "tb2_read %s: format='%s' is too long(>%d) for tbl='%s'\n",
	    TBL_VER, buff+ipos[0]+1, L_FORMAT, filename );
#endif
    tb2_errno = TBL_E_KEYWORDS;
    return (void *)NULL;

 EPOSITION:  /* message appear in _tbl_position() */
    tb2_errno = TBL_E_KEYWORDS;
    return (void *)NULL;

 ELINE:      /* message appear in the loop */
    tb2_errno = TBL_E_ILLLINE;
    return (void *)NULL;

 EMALLOC:
#ifdef TBL_WARN
    fprintf( WARN_DEST,
	    "tb2_read %s: memory allocation failure\n", 
	    TBL_VER );
#endif
    tb2_errno = TBL_E_BUFFSIZE;
    return (void *)NULL;

ENULLTBL:
#ifdef TBL_WARN
    fprintf( WARN_DEST,
	    "tb2_read %s: table = '%s' contains no data \n", 
	    TBL_VER, filename );
#endif
    tb2_errno = TBL_E_NULLTBL;
    return (void *)NULL;
}

/**
 *  int _tbl_position -- get separator position(s) in a string
 *
 *  return: # of 'separator's in the specified string (0,1,2,..)
**/
int _tbl_position( str, separator, ipos )
char  *str;        /* string to analyze */
char  separator;   /* char.=separator to be searched */
int   *ipos;       /* result: positions in the specified string */
{
  int  i, counts,len;

  counts = 0;*ipos=0;
  len = strlen( str );
    if( len <= 0 )  return 0;;

  for( i=0 ; i<len ; ++i ) {
    if( *(str+i) == separator ) {
      *(ipos+counts) = i;
      if( counts >= N_POS ) {
#ifdef TBL_WARN
      fprintf( WARN_DEST,
	"tb2_read %s: too many (>%d) separators='%c' found in the string\n", 
        TBL_VER, N_POS, separator );
      fprintf( WARN_DEST,
	  "          --> '%s'\n", str );
#endif
	return (-1);
      }
      counts++;
    }
  }
  return  counts;
}

/**
 *   tbl_read() - read specified table into buffer (old interface)
**/
int tbl_read( abbr )
char  *abbr;			/* abbr. name of the table */
{
    default_tbl = tb2_read( abbr );		/* call version 2 library */
    if ( default_tbl )  {
	return 0;
    }  else  {
	return tb2_errno;
    }
}

/*
 * _tb2_strdupaddn  --  duplicate string with "%n" appended for sscanf format
 */
static char * _tb2_strdupaddn( char * format )
{
    char *addn;
    unsigned int len;

    if (!format)  return((char *)NULL);
    len = strlen(format) + 3;
    if (!(addn = tb2_malloc( len )))  return((char *)NULL);
    strncpy(addn, format, len);
    strncat(addn, "%n\0", 3);
    return addn;
}
